# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils for common IO operations over archive files, e.g. tar or gzip."""

from __future__ import annotations

__all__ = [
    "SUPPORTED_ARCHIVE_FORMATS",
    "open_gz_file",
    "extract_file",
]

from contextlib import contextmanager
import gzip
from pathlib import Path
import shutil
from typing import Generator, TextIO

from ensembl.utils import StrPath
from ensembl.utils.argparse import ArgumentParser


def _unpack_gz_files(src_file: StrPath, dst_dir: StrPath) -> None:
    """Unpacks `src_file` to `dst_dir`.

    Args:
        src_file: File path to unpack (with ".gz" extension).
        dst_dir: Directory path to unpack the file into.

    """
    # Remove '.gz' extension to create the destination file name
    dst_file = Path(dst_dir) / Path(src_file).stem
    with gzip.open(src_file, "rb") as f_in:
        with dst_file.open("wb") as f_out:
            shutil.copyfileobj(f_in, f_out)


shutil.register_unpack_format("gzip", [".gz"], _unpack_gz_files, description="GZIP file")

# Each registered format is a tuple (name, extensions, description)
SUPPORTED_ARCHIVE_FORMATS = [ext for elem in shutil.get_unpack_formats() for ext in elem[1]]


@contextmanager
def open_gz_file(file_path: StrPath) -> Generator[TextIO, None, None]:
    """Yields an open file object, even if the file is compressed with gzip.

    The file is expected to contain a text, and this can be used with the usual "with".

    Args:
        file_path: A (single) file path to open.

    """
    src_file = Path(file_path)
    if src_file.suffix == ".gz":
        with gzip.open(src_file, "rt") as fh:
            yield fh
    else:
        with src_file.open("rt") as fh:
            yield fh


def extract_file(src_file: StrPath, dst_dir: StrPath) -> None:
    """Extracts the `src_file` into `dst_dir`.

    If the file is not an archive, it will be copied to `dst_dir`. `dst_dir` will be created if it
    does not exist.

    Args:
        src_file: Path to the file to unpack.
        dst_dir: Path to the folder where to extract the file.

    """
    src_file = Path(src_file)
    extensions = {"".join(src_file.suffixes[i:]) for i in range(0, len(src_file.suffixes))}

    if extensions.intersection(SUPPORTED_ARCHIVE_FORMATS):
        shutil.unpack_archive(src_file, dst_dir)
    else:
        # Replicate the functionality of shutil.unpack_archive() by creating `dst_dir`
        Path(dst_dir).mkdir(parents=True, exist_ok=True)
        shutil.copy(src_file, dst_dir)


def extract_file_cli() -> None:
    """Entry-point for the `extract_file` method"""
    parser = ArgumentParser(description="Extracts file to the given location.")
    parser.add_argument_src_path("--src_file", required=True, help="Path to the file to unpack")
    parser.add_argument_dst_path(
        "--dst_dir", default=Path.cwd(), help="Path to the folder where to extract the file"
    )
    args = parser.parse_args()
    extract_file(args.src_file, args.dst_dir)



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Database connection handler.

This module provides the main class to connect to and access databases. It will be an ORM-less
connection, that is, the data can only be accessed via SQL queries (see example below).

Examples:

    >>> from ensembl.utils.database import DBConnection
    >>> dbc = DBConnection("mysql://ensro@mysql-server:4242/mydb")
    >>> # You can access the database data via sql queries, for instance:
    >>> results = dbc.execute("SELECT * FROM my_table;")
    >>> # Or via a connection in a transaction manner:
    >>> with dbc.begin() as conn:
    >>>     results = conn.execute("SELECT * FROM my_table;")

"""

from __future__ import annotations

__all__ = [
    "Query",
    "StrURL",
    "DBConnection",
]

from contextlib import contextmanager
from typing import ContextManager, Generator, Optional, TypeVar

import sqlalchemy
from sqlalchemy import create_engine, event, text
from sqlalchemy.orm import sessionmaker


Query = TypeVar("Query", str, sqlalchemy.sql.expression.ClauseElement, sqlalchemy.sql.expression.TextClause)
StrURL = TypeVar("StrURL", str, sqlalchemy.engine.URL)


class DBConnection:
    """Database connection handler, providing also the database's schema and properties.

    Args:
        url: URL to the database, e.g. `mysql://user:passwd@host:port/my_db`.

    """

    def __init__(self, url: StrURL, **kwargs) -> None:
        self._engine = create_engine(url, future=True, **kwargs)
        self.load_metadata()

    def __repr__(self) -> str:
        """Returns a string representation of this object."""
        return f"{self.__class__.__name__}({self.url!r})"

    def load_metadata(self) -> None:
        """Loads the metadata information of the database."""
        # Note: Just reflect() is not enough as it would not delete tables that no longer exist
        self._metadata = sqlalchemy.MetaData()
        self._metadata.reflect(bind=self._engine)

    @property
    def url(self) -> str:
        """Returns the database URL."""
        return self._engine.url.render_as_string(hide_password=False)

    @property
    def db_name(self) -> Optional[str]:
        """Returns the database name."""
        return self._engine.url.database

    @property
    def host(self) -> Optional[str]:
        """Returns the database host."""
        return self._engine.url.host

    @property
    def port(self) -> Optional[int]:
        """Returns the port of the database host."""
        return self._engine.url.port

    @property
    def dialect(self) -> str:
        """Returns the SQLAlchemy database dialect name of the database host."""
        return self._engine.name

    @property
    def tables(self) -> dict[str, sqlalchemy.schema.Table]:
        """Returns the database tables keyed to their name."""
        return self._metadata.tables

    def get_primary_key_columns(self, table: str) -> list[str]:
        """Returns the primary key column names for the given table.

        Args:
            table: Table name.

        """
        return [col.name for col in self.tables[table].primary_key.columns.values()]

    def get_columns(self, table: str) -> list[str]:
        """Returns the column names for the given table.

        Args:
            table: Table name.

        """
        return [col.name for col in self.tables[table].columns]

    def connect(self) -> sqlalchemy.engine.Connection:
        """Returns a new database connection."""
        return self._engine.connect()

    def begin(self, *args) -> ContextManager[sqlalchemy.engine.Connection]:
        """Returns a context manager delivering a database connection with a transaction established."""
        return self._engine.begin(*args)

    def dispose(self) -> None:
        """Disposes of the connection pool."""
        self._engine.dispose()

    def execute(self, statement: Query, parameters=None, execution_options=None) -> sqlalchemy.engine.Result:
        """Executes the given SQL query and returns its result.

        See `sqlalchemy.engine.Connection.execute()` method for more information about the
        additional arguments.

        Args:
            statement: SQL query to execute.
            parameters: Parameters which will be bound into the statement.
            execution_options: Optional dictionary of execution options, which will be associated
                with the statement execution.

        """
        if isinstance(statement, str):
            statement = text(statement)  # type: ignore[assignment]
        return self.connect().execute(
            statement=statement, parameters=parameters, execution_options=execution_options
        )  # type: ignore[call-overload]

    def _enable_sqlite_savepoints(self, engine: sqlalchemy.engine.Engine) -> None:
        """Enables SQLite SAVEPOINTS to allow session rollbacks."""

        @event.listens_for(engine, "connect")
        def do_connect(dbapi_connection, connection_record):  # pylint: disable=unused-argument
            """Disables emitting the BEGIN statement entirely, as well as COMMIT before any DDL."""
            dbapi_connection.isolation_level = None

        @event.listens_for(engine, "begin")
        def do_begin(conn):
            """Emits a customour own BEGIN."""
            conn.exec_driver_sql("BEGIN")

    @contextmanager
    def session_scope(self) -> Generator[sqlalchemy.orm.session.Session, None, None]:
        """Provides a transactional scope around a series of operations with rollback in case of failure.

        Bear in mind MySQL's storage engine MyISAM does not support rollback transactions, so all
        the modifications performed to the database will persist.

        """
        # Create a dedicated engine for this session
        engine = create_engine(self._engine.url)
        if self.dialect == "sqlite":
            self._enable_sqlite_savepoints(engine)
        Session = sessionmaker(future=True)
        session = Session(bind=engine, autoflush=False)
        try:
            yield session
            session.commit()
        except:
            # Rollback to ensure no changes are made to the database
            session.rollback()
            raise
        finally:
            # Whatever happens, make sure the session is closed
            session.close()

    @contextmanager
    def test_session_scope(self) -> Generator[sqlalchemy.orm.session.Session, None, None]:
        """Provides a transactional scope around a series of operations that will be rolled back at the end.

        Bear in mind MySQL's storage engine MyISAM does not support rollback transactions, so all
        the modifications performed to the database will persist.

        """
        # Create a dedicated engine for this session
        engine = create_engine(self._engine.url)
        if self.dialect == "sqlite":
            self._enable_sqlite_savepoints(engine)
        # Connect to the database
        connection = engine.connect()
        # Begin a non-ORM transaction
        transaction = connection.begin()
        # Bind an individual session to the connection
        Session = sessionmaker(future=True)
        try:
            # Running on SQLAlchemy 2.0+
            session = Session(bind=connection, join_transaction_mode="create_savepoint")
        except TypeError:
            # Running on SQLAlchemy 1.4
            session = Session(bind=connection)
            # If the database supports SAVEPOINT, starting a savepoint will allow to also use rollback
            connection.begin_nested()

            # Define a new transaction event
            @event.listens_for(session, "after_transaction_end")
            def end_savepoint(session, transaction):  # pylint: disable=unused-argument
                if not connection.in_nested_transaction():
                    connection.begin_nested()

        try:
            yield session
        finally:
            # Whatever happens, make sure the session and connection are closed, rolling back
            # everything done with the session (including calls to commit())
            session.close()
            transaction.rollback()
            connection.close()



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Easy initialisation functionality to set an event logging system.

Examples:

    >>> import logging, pathlib
    >>> from ensembl.utils.logging import init_logging
    >>> logfile = pathlib.Path("test.log")
    >>> init_logging("INFO", logfile, "DEBUG")
    >>> logging.info("This message is written in both stderr and the log file")
    >>> logging.debug("This message is only written in the log file")

"""

from __future__ import annotations

__all__ = [
    "LogLevel",
    "init_logging",
    "init_logging_with_args",
]

import argparse
import logging
from typing import Optional, Union

from ensembl.utils import StrPath


LogLevel = Union[int, str]


def init_logging(
    log_level: LogLevel = "WARNING",
    log_file: Optional[StrPath] = None,
    log_file_level: LogLevel = "DEBUG",
    msg_format: str = "%(asctime)s\t%(levelname)s\t%(message)s",
    date_format: str = r"%Y-%m-%d_%H:%M:%S",
) -> None:
    """Initialises the logging system.

    By default, all the log messages corresponding to `log_level` (and above) will be printed in the
    standard error. If `log_file` is provided, all messages of `log_file_level` level (and above) will
    be written into the provided file.

    Args:
        log_level: Minimum logging level for the standard error.
        log_file: Logging file where to write logging messages besides the standard error.
        log_file_level: Minimum logging level for the logging file.
        msg_format: A format string for the logged output as a whole. More information:
            https://docs.python.org/3/library/logging.html#logrecord-attributes
        date_format: A format string for the date/time portion of the logged output. More information:
            https://docs.python.org/3/library/logging.html#logging.Formatter.formatTime

    """
    # Configure the basic logging system, setting the root logger to the minimum log level available
    # to avoid filtering messages in any handler due to "parent delegation". Also close and remove any
    # existing handlers before setting this configuration.
    logging.basicConfig(format=msg_format, datefmt=date_format, level="DEBUG", force=True)
    # Set the correct log level of the new StreamHandler (by default it is set to NOTSET)
    logging.root.handlers[0].setLevel(log_level)
    if log_file:
        # Create the log file handler and add it to the root logger
        formatter = logging.Formatter(msg_format, datefmt=date_format)
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(log_file_level)
        file_handler.setFormatter(formatter)
        logging.root.addHandler(file_handler)


def init_logging_with_args(args: argparse.Namespace) -> None:
    """Processes the Namespace object provided to call `init_logging()` with the correct arguments.

    Args:
        args: Namespace populated by an argument parser.

    """
    args_dict = vars(args)
    log_args = {x: args_dict[x] for x in ["log_level", "log_file", "log_file_level"] if x in args_dict}
    init_logging(**log_args)



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.archive` module."""

import filecmp
from pathlib import Path

import pytest
from pytest import param

from ensembl.utils.archive import open_gz_file, extract_file


@pytest.mark.parametrize(
    "src_file, expected_file",
    [
        param("sample.txt.gz", "sample.txt", id="gzip file"),
        param("sample.txt", "sample.txt", id="uncompressed file"),
    ],
)
def test_open_gz_file(data_dir, src_file, expected_file) -> None:
    """Tests `open_gz_file()` function.

    Args:
        tmp_path: Fixture that provides a temporary directory path unique to the test invocation.
        data_dir: Fixture that provides the path to the test data folder matching the test's name.
        src_file: Source file to open.
        expected_file: File with the expected content to be found in the source file.

    """
    # Get content of archive file
    with open_gz_file(data_dir / src_file) as in_file:
        src_content = in_file.readlines()
    # Get expected content
    expected_path = data_dir / expected_file
    with expected_path.open("r") as in_file:
        expected_content = in_file.readlines()
    assert src_content == expected_content


@pytest.mark.parametrize(
    "src_file, expected_output",
    [
        param("sample.tar", "sample.txt", id="tar file"),
        param("sample.txt.gz", "sample.txt", id="gzip file"),
        param("sample.txt", "sample.txt", id="uncompressed file"),
    ],
)
def test_extract_file(tmp_path: Path, data_dir: Path, src_file: str, expected_output: str) -> None:
    """Tests `extract_file()` function.

    Args:
        tmp_path: Fixture that provides a temporary directory path unique to the test invocation.
        data_dir: Fixture that provides the path to the test data folder matching the test's name.
        src_file: Source file to extract.
        expected_output: File with the expected content after extracting the source file.

    """
    extract_file(data_dir / src_file, tmp_path)
    assert filecmp.cmp(tmp_path / expected_output, data_dir / expected_output)



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate the code reference pages for mkdocs."""

from pathlib import Path

import mkdocs_gen_files


nav = mkdocs_gen_files.Nav()

root = Path("src").resolve()
for py_path in sorted(root.rglob("*.py")):
    # Get the relative module path and corresponding documentation paths
    module_path = py_path.relative_to(root)
    doc_path = module_path.with_suffix(".md")
    full_doc_path = Path("reference", doc_path)
    # Get all the parts of the module path without the ".py" extension
    parts = tuple(module_path.with_suffix("").parts)
    # Drop "__init__" file from the path components as well (if present)
    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    # Add markdown file path with its index tree
    nav[parts] = doc_path.as_posix()
    # Populate the markdown file with the doc stub of this module
    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        identifier = ".".join(parts)
        fd.write(f"::: {identifier}\n")
    # Correct the path
    mkdocs_gen_files.set_edit_path(full_doc_path, module_path)

with mkdocs_gen_files.open("reference/summary.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ensembl's pytest plugin with useful unit testing hooks and fixtures."""

from __future__ import annotations

from difflib import unified_diff
import os
from pathlib import Path
import re
from typing import Callable, Generator, Optional

import pytest
from pytest import Config, FixtureRequest, Parser

from ensembl.utils import StrPath
from ensembl.utils.database import UnitTestDB


def pytest_addoption(parser: Parser) -> None:
    """Registers argparse-style options for Ensembl's unit testing.

    `Pytest initialisation hook
    <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_addoption>`_.

    Args:
        parser: Parser for command line arguments and ini-file values.

    """
    # Add the Ensembl unitary test parameters to pytest parser
    group = parser.getgroup("Ensembl unit testing")
    group.addoption(
        "--server",
        action="store",
        metavar="URL",
        dest="server",
        required=False,
        default=os.getenv("DB_HOST", "sqlite:///"),
        help="Server URL where to create the test database(s)",
    )
    group.addoption(
        "--keep-dbs",
        action="store_true",
        dest="keep_dbs",
        required=False,
        help="Do not remove the test databases (default: False)",
    )


def pytest_report_header(config: Config) -> str:
    """Presents extra information in the report header.

    Args:
        config: Access to configuration values, pluginmanager and plugin hooks.

    """
    # Show server information, masking the password value
    server = config.getoption("server")
    server = re.sub(r"(//[^/]+:).*(@)", r"\1xxxxxx\2", server)
    return f"server: {server}"


@pytest.fixture(name="data_dir", scope="module")
def local_data_dir(request: FixtureRequest) -> Path:
    """Returns the path to the test data folder matching the test's name.

    Args:
        request: Fixture that provides information of the requesting test function.

    """
    return Path(request.module.__file__).with_suffix("")


@pytest.fixture(name="assert_files")
def fixture_assert_files() -> Callable[[StrPath, StrPath], None]:
    """Returns a function that asserts if two text files are equal, or prints their differences."""

    def _assert_files(result_path: StrPath, expected_path: StrPath) -> None:
        """Asserts if two files are equal, or prints their differences.

        Args:
            result_path: Path to results (test-made) file.
            expected_path: Path to expected file.

        """
        with open(result_path, "r") as result_fh:
            results = result_fh.readlines()
        with open(expected_path, "r") as expected_fh:
            expected = expected_fh.readlines()
        files_diff = list(
            unified_diff(
                results,
                expected,
                fromfile=f"Test-made file {Path(result_path).name}",
                tofile=f"Expected file {Path(expected_path).name}",
            )
        )
        assert_message = f"Test-made and expected files differ\n{' '.join(files_diff)}"
        assert len(files_diff) == 0, assert_message

    return _assert_files


@pytest.fixture(name="db_factory", scope="module")
def fixture_db_factory(request: FixtureRequest, data_dir: Path) -> Generator[Callable, None, None]:
    """Yields a unit test database factory.

    Args:
        request: Fixture that provides information of the requesting test function.
        data_dir: Fixture that provides the path to the test data folder matching the test's name.

    """
    created: dict[str, UnitTestDB] = {}
    server_url = request.config.getoption("server")

    def _db_factory(src: StrPath, name: Optional[str] = None) -> UnitTestDB:
        """Returns a unit test database.

        Args:
            src: Directory path where the test database schema and content files are located.
            name: Name to give to the new database. See `UnitTestDB` for more information.

        """
        src_path = Path(src)
        if not src_path.is_absolute():
            src_path = data_dir / src_path
        db_key = name if name else src_path.name
        return created.setdefault(db_key, UnitTestDB(server_url, src_path, name))

    yield _db_factory
    # Drop all unit test databases unless the user has requested to keep them
    if not request.config.getoption("keep_dbs"):
        for test_db in created.values():
            test_db.drop()


@pytest.fixture(scope="module")
def test_dbs(request: FixtureRequest, db_factory: Callable) -> dict[str, UnitTestDB]:
    """Returns a dictionary of unit test databases with the database name as key.

    Requires a list of dictionaries, each with keys `src` (mandatory) and `name` (optional), passed via
    `request.param`. See `db_factory()` for details about each key's value. This fixture is a wrapper of
    `db_factory()` intended to be used via indirect parametrization, for example::

        @pytest.mark.parametrize(
            "test_dbs", [[{"src": "master"}, {"src": "master", "name": "master2"}]], indirect=True
        )
        def test_method(..., test_dbs: dict[str, UnitTestDB], ...):


    Args:
        request: Fixture that provides information of the requesting test function.
        db_factory: Fixture that provides a unit test database factory.

    """
    databases = {}
    for argument in request.param:
        src = Path(argument["src"])
        name = argument.get("name", None)
        key = name if name else src.name
        databases[key] = db_factory(src, name)
    return databases



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provide an extended version of `argparse.ArgumentParser` with additional functionality.

Examples:

    >>> from pathlib import Path
    >>> from ensembl.util.argparse import ArgumentParser
    >>> parser = ArgumentParser(description="Tool description")
    >>> parser.add_argument_src_path("--src_file", required=True, help="Path to source file")
    >>> parser.add_server_arguments(help="Server to connect to")
    >>> args = parser.parse_args()
    >>> args
    Namespace(host='myserver', port=3826, src_file=PosixPath('/path/to/src_file.txt'),
    url=URL('mysql://username@myserver:3826'), user='username')

"""

from __future__ import annotations

__all__ = [
    "ArgumentParser",
]

import argparse
import os
from pathlib import Path

from sqlalchemy.engine import make_url, URL

from ensembl.utils import StrPath


class ArgumentParser(argparse.ArgumentParser):
    """Extends `argparse.ArgumentParser` with additional methods and functionality.

    The default behaviour of the help text will be to display the default values on every non-required
    argument, i.e. optional arguments with `required=False`.

    """

    def __init__(self, *args, **kwargs) -> None:
        """Extends the base class to include the information about default argument values by default."""
        super().__init__(*args, **kwargs)
        self.formatter_class = argparse.ArgumentDefaultsHelpFormatter
        self.__server_groups: list[str] = []

    def _validate_src_path(self, src_path: StrPath) -> Path:
        """Returns the path if exists and it is readable, raises an error through the parser otherwise.

        Args:
            src_path: File or directory path to check.

        """
        src_path = Path(src_path)
        if not src_path.exists():
            self.error(f"'{src_path}' not found")
        elif not os.access(src_path, os.R_OK):
            self.error(f"'{src_path}' not readable")
        return src_path

    def _validate_dst_path(self, dst_path: StrPath, exists_ok: bool = False) -> Path:
        """Returns the path if it is writable, raises an error through the parser otherwise.

        Args:
            dst_path: File or directory path to check.
            exists_ok: Do not raise an error during parsing if the destination path already exists.

        """
        dst_path = Path(dst_path)
        if dst_path.exists():
            if os.access(dst_path, os.W_OK):
                if exists_ok:
                    return dst_path
                self.error(f"'{dst_path}' already exists")
            else:
                self.error(f"'{dst_path}' is not writable")
        # Check if the first parent directory that exists is writable
        for parent_path in dst_path.parents:
            if parent_path.exists():
                if not os.access(parent_path, os.W_OK):
                    self.error(f"'{dst_path}' is not writable")
                break
        return dst_path

    def add_argument(self, *args, **kwargs) -> None:  # type: ignore[override]
        """Extends the parent function by excluding the default value in the help text when not provided.

        Only applied to required arguments without a default value, i.e. positional arguments or optional
        arguments with `required=True`.

        """
        if kwargs.get("required", False):
            kwargs.setdefault("default", argparse.SUPPRESS)
        super().add_argument(*args, **kwargs)

    def add_argument_src_path(self, *args, **kwargs) -> None:
        """Adds `pathlib.Path` argument, checking if it exists and it is readable at parsing time.

        If "metavar" is not defined, it is added with "PATH" as value to improve help text readability.

        """
        kwargs.setdefault("metavar", "PATH")
        kwargs["type"] = self._validate_src_path
        self.add_argument(*args, **kwargs)

    def add_argument_dst_path(self, *args, exists_ok: bool = True, **kwargs) -> None:
        """Adds `pathlib.Path` argument, checking if it is writable at parsing time.

        If "metavar" is not defined it is added with "PATH" as value to improve help text readability.

        Args:
            exists_ok: Do not raise an error if the destination path already exists.

        """
        kwargs.setdefault("metavar", "PATH")
        kwargs["type"] = lambda x: self._validate_dst_path(x, exists_ok)
        self.add_argument(*args, **kwargs)

    def add_argument_url(self, *args, **kwargs) -> None:
        """Adds `sqlalchemy.engine.URL` argument.

        If "metavar" is not defined it is added with "URI" as value to improve help text readability.

        """
        kwargs.setdefault("metavar", "URI")
        kwargs["type"] = make_url
        self.add_argument(*args, **kwargs)

    # pylint: disable=redefined-builtin
    def add_server_arguments(self, prefix: str = "", include_database: bool = False, help: str = "") -> None:
        """Adds the usual set of arguments needed to connect to a server, i.e. `--host`, `--port`, `--user`
        and `--password` (optional).

        Note that the parser will assume this is a MySQL server.

        Args:
            prefix: Prefix to add the each argument, e.g. if prefix is `src_`, the arguments will be
                `--src_host`, etc.
            include_database: Include `--database` argument.
            help: Description message to include for this set of arguments.

        """
        group = self.add_argument_group(f"{prefix}server connection arguments", description=help)
        group.add_argument(f"--{prefix}host", required=True, metavar="HOST", help="host name")
        group.add_argument(f"--{prefix}port", required=True, type=int, metavar="PORT", help="port number")
        group.add_argument(f"--{prefix}user", required=True, metavar="USER", help="user name")
        group.add_argument(f"--{prefix}password", metavar="PWD", help="host password")
        if include_database:
            group.add_argument(f"--{prefix}database", required=True, metavar="NAME", help="database name")
        self.__server_groups.append(prefix)

    def add_log_arguments(self, add_log_file: bool = False) -> None:
        """Adds the usual set of arguments required to set and initialise a logging system.

        The current set includes a mutually exclusive group for the default logging level: `--verbose`,
        `--debug` or `--log LEVEL`.

        Args:
            add_log_file: Add arguments to allow storing messages into a file, i.e. `--log_file` and
                `--log_file_level`.

        """
        # Define the list of log levels available
        log_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        # NOTE: from 3.11 this list can be changed to: logging.getLevelNamesMapping().keys()
        # Create logging arguments group
        group = self.add_argument_group("logging arguments")
        # Add 3 mutually exclusive options to set the logging level
        subgroup = group.add_mutually_exclusive_group()
        subgroup.add_argument(
            "-v",
            "--verbose",
            action="store_const",
            const="INFO",
            dest="log_level",
            help="verbose mode, i.e. 'INFO' log level",
        )
        subgroup.add_argument(
            "--debug",
            action="store_const",
            const="DEBUG",
            dest="log_level",
            help="debugging mode, i.e. 'DEBUG' log level",
        )
        subgroup.add_argument(
            "--log",
            choices=log_levels,
            type=str.upper,
            default="WARNING",
            metavar="LEVEL",
            dest="log_level",
            help="level of the events to track: %(choices)s",
        )
        subgroup.set_defaults(log_level="WARNING")
        if add_log_file:
            # Add log file-related arguments
            group.add_argument(
                "--log_file",
                type=lambda x: self._validate_dst_path(x, exists_ok=True),
                metavar="PATH",
                default=None,
                help="log file path",
            )
            group.add_argument(
                "--log_file_level",
                choices=log_levels,
                type=str.upper,
                default="DEBUG",
                metavar="LEVEL",
                help="level of the events to track in the log file: %(choices)s",
            )

    def parse_args(self, *args, **kwargs) -> argparse.Namespace:  # type: ignore[override]
        """Extends the parent function by adding a new URL argument for every server group added.

        The type of this new argument will be `sqlalchemy.engine.URL`. It also logs all the parsed
        arguments for debugging purposes when logging arguments have been added.

        """
        arguments = super().parse_args(*args, **kwargs)
        # Build and add an sqlalchemy.engine.URL object for every server group added
        for prefix in self.__server_groups:
            # Raise an error rather than overwriting when the URL argument is already present
            if f"{prefix}url" in arguments:
                self.error(f"argument '{prefix}url' is already present")
            server_url = URL.create(
                "mysql",
                getattr(arguments, f"{prefix}user"),
                getattr(arguments, f"{prefix}password"),
                getattr(arguments, f"{prefix}host"),
                getattr(arguments, f"{prefix}port"),
                getattr(arguments, f"{prefix}database", None),
            )
            setattr(arguments, f"{prefix}url", server_url)
        return arguments



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.plugin` module.

Since certain elements are embedded within pytest itself, only the fixtures are unit tested in this case.
"""

from contextlib import nullcontext as does_not_raise
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, ContextManager
from unittest.mock import patch

import pytest
from pytest import FixtureRequest, param, raises

from ensembl.utils import StrPath
from ensembl.utils.database import StrURL


@dataclass
class MockTestDB:
    """Mocks `UnitTestDB` class by just storing the three arguments provided."""

    server_url: StrURL
    dump_dir: StrPath
    name: str

    def drop(self) -> None:
        """Mocks `UnitTestDB.drop()` method."""


@pytest.mark.dependency(name="test_data_dir")
def test_data_dir(request: FixtureRequest, data_dir: Path) -> None:
    """Tests the `data_dir` fixture.

    Args:
        request: Fixture that provides information of the requesting test function.
        data_dir: Fixture that provides the path to the test data folder matching the test's name.

    """
    assert data_dir.stem == request.path.stem


@pytest.mark.dependency(depends=["test_data_dir"])
@pytest.mark.parametrize(
    "left, right, expectation",
    [
        param("file1.txt", "file1.txt", does_not_raise(), id="Files are equal"),
        param("file1.txt", "file2.txt", raises(AssertionError), id="Files differ"),
    ],
)
def test_assert_files(
    assert_files: Callable, data_dir: Path, left: str, right: str, expectation: ContextManager
) -> None:
    """Tests the `assert_files` fixture.

    Args:
        assert_files: Fixture that provides an assertion function to compare two files.
        data_dir: Fixture that provides the path to the test data folder matching the test's name.
        left: Left file to compare.
        right: Right file to compare.
        expectation: Context manager for the expected exception.

    """
    with expectation:
        assert_files(data_dir / left, data_dir / right)


@pytest.mark.parametrize(
    "dump_dir, db_name",
    [
        (Path("dump_dir"), "dump_dir"),
        (Path("dump_dir").resolve(), "test_db"),
    ],
)
@patch("ensembl.utils.plugin.UnitTestDB", new=MockTestDB)
def test_db_factory(request: FixtureRequest, db_factory: Callable, dump_dir: Path, db_name: str) -> None:
    """Tests the `db_factory` fixture.

    Args:
        request: Fixture that provides information of the requesting test function.
        db_factory: Fixture that provides a unit test database factory.
        dump_dir: Directory path where the test database schema and content files are located.
        db_name: Name to give to the new database.

    """
    test_db = db_factory(dump_dir, db_name)
    assert test_db.server_url == request.config.getoption("server")
    assert test_db.name == db_name
    if dump_dir.is_absolute():
        assert test_db.dump_dir == dump_dir
    else:
        assert test_db.dump_dir.stem == str(dump_dir)



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.argparse` module."""

from contextlib import nullcontext as does_not_raise
from pathlib import Path
import re
from typing import Any, ContextManager

import pytest
from pytest import param, raises
from sqlalchemy.engine import make_url

from ensembl.utils.argparse import ArgumentParser


def _args_dict_to_cmd_list(args_dict: dict[str, Any]) -> list[str]:
    """Returns a flattened version of the arguments dictionary in a list with every element as a string.

    It will add "--" to every key of the dictionary, and will not add empty values (for flag arguments).

    """
    cmd_args = []
    for key, value in args_dict.items():
        cmd_args.append(f"--{key}")
        if value:
            cmd_args.append(str(value))
    return cmd_args


class TestArgumentParser:
    """Tests `ArgumentParser` class."""

    def test_validate_unreadable_src_path(self, tmp_path: Path) -> None:
        """Tests `ArgumentParser._validate_src_path()` method for an unreadable file.

        Args:
            tmp_path: Fixture that provides a temporary directory path unique to the test invocation.

        """
        # Create file and make it unreadable
        src_path = tmp_path / "forbidden.txt"
        with src_path.open("w") as in_file:
            in_file.write("you shall not read this")
        src_path.chmod(0o222)
        # Attempt to validate source path
        parser = ArgumentParser()
        with raises(SystemExit):
            parser._validate_src_path(src_path)  # pylint: disable=protected-access

    @pytest.mark.parametrize(
        "is_dir",
        [
            param(True, id="Destination path is a file"),
            param(False, id="Destination path is a directory"),
        ],
    )
    def test_validate_unwritable_dst_path(self, tmp_path: Path, is_dir: bool) -> None:
        """Tests `ArgumentParser._validate_dst_path()` method for an unwritable path.

        Args:
            tmp_path: Fixture that provides a temporary directory path unique to the test invocation.
            is_dir: Make the unwritable destination path a directory (`True`) or file (`False`).

        """
        # Create a file or directory and make it unreadable
        if is_dir:
            dst_path = tmp_path / "read-only" / "read-only.txt"
            dst_path.parent.mkdir(mode=0o555)
        else:
            dst_path = tmp_path / "read-only.txt"
            with dst_path.open("w") as in_file:
                in_file.write("you shall not edit this")
            dst_path.chmod(0o444)
        # Attempt to validate source path
        parser = ArgumentParser()
        with raises(SystemExit):
            parser._validate_dst_path(dst_path)  # pylint: disable=protected-access

    @pytest.mark.dependency(name="add_argument")
    @pytest.mark.parametrize(
        "required",
        [
            param(True, id="Add required argument"),
            param(False, id="Add optional argument"),
        ],
    )
    def test_add_argument(self, required: bool) -> None:
        """Tests `ArgumentParser.add_argument()` method.

        Args:
            required: Set argument as required (`True`) or optional (`False`).

        """
        parser = ArgumentParser()
        parser.add_argument("--foo", help="text", required=required)
        if required:
            pattern = re.compile(r" +--foo FOO +text")
        else:
            pattern = re.compile(r" +--foo FOO +text \(default: None\)")
        assert pattern.search(parser.format_help()) is not None

    @pytest.mark.dependency(depends=["add_argument"])
    @pytest.mark.parametrize(
        "src_path, expectation",
        [
            param("", does_not_raise(), id="Optional input file"),
            param("sample.txt", does_not_raise(), id="Input file exists"),
            param("invalid.txt", raises(SystemExit), id="Input file does not exist"),
        ],
    )
    def test_add_argument_src_path(self, data_dir: Path, src_path: str, expectation: ContextManager) -> None:
        """Tests `ArgumentParser.add_argument_src_path()` method.

        Args:
            data_dir: Fixture that provides the path to the test data folder matching the test's name.
            src_path: Input file name.
            expectation: Context manager for the expected exception.

        """
        # Add "--src_path" argument to the parser and its command line
        parser = ArgumentParser()
        parser.add_argument_src_path("--src_path")
        cmd_args = ["--src_path", str(data_dir / src_path)]
        # Check that the argument is properly parsed
        with expectation:
            args = parser.parse_args(cmd_args)
            assert args.src_path == data_dir / src_path

    @pytest.mark.dependency(depends=["add_argument"])
    @pytest.mark.parametrize(
        "dst_path, exists_ok, expectation",
        [
            param("", True, does_not_raise(), id="Dir exists and writable, exists_ok = True"),
            param("", False, raises(SystemExit), id="Dir exists and writable, exists_ok = False"),
            param("output", True, does_not_raise(), id="Dir does not exist but parent is writable"),
            param("dir1/dir2", True, does_not_raise(), id="Dir does not exist but grandparent is writable"),
        ],
    )
    def test_add_argument_dst_path(
        self, data_dir: Path, dst_path: str, exists_ok: bool, expectation: ContextManager
    ) -> None:
        """Tests `ArgumentParser.add_argument_dst_path()` method.

        Args:
            data_dir: Fixture that provides the path to the test data folder matching the test's name.
            dst_path: Relative path to destination directory/file.
            exists_ok: Do not raise an error if the destination path already exists.
            expectation: Context manager for the expected exception.

        """
        # Add "--dst_path" argument to the parser and its command line
        parser = ArgumentParser()
        parser.add_argument_dst_path("--dst_path", exists_ok=exists_ok)
        cmd_args = ["--dst_path", str(data_dir / dst_path)]
        # Check that the argument is properly parsed
        with expectation:
            args = parser.parse_args(cmd_args)
            assert args.dst_path == data_dir / dst_path

    @pytest.mark.dependency(depends=["add_argument"])
    def test_add_argument_url(self) -> None:
        """Tests `ArgumentParser.add_argument_url()` method."""
        # Add "--url" argument to the parser and its command line
        parser = ArgumentParser()
        parser.add_argument_url("--url")
        cmd_args = ["--url", "https://github.com"]
        # Check that the argument is properly parsed
        args = parser.parse_args(cmd_args)
        assert args.url == make_url("https://github.com")

    @pytest.mark.dependency(name="add_server_arguments", depends=["add_argument"])
    @pytest.mark.parametrize(
        "prefix, include_database",
        [
            param("", False, id="Basic call"),
            param("src", False, id="Add prefix"),
            param("", True, id="Add database argument"),
        ],
    )
    def test_add_server_arguments(self, prefix: str, include_database: bool) -> None:
        """Tests `ArgumentParser.add_server_arguments()` method.

        Args:
            prefix: Prefix to add the each argument.
            include_database: Include `--database` argument.

        """
        # Add server arguments to the parser and its command line
        parser = ArgumentParser()
        parser.add_server_arguments(prefix=prefix, include_database=include_database)
        cmd_args = {
            f"{prefix}host": "lugh",
            f"{prefix}port": 42,
            f"{prefix}user": "username",
            f"{prefix}password": "***",
        }
        url = "mysql://username:***@lugh:42"
        if include_database:
            cmd_args[f"{prefix}database"] = "my_db"
            url += "/my_db"
        # Check that the arguments are properly parsed
        args = parser.parse_args(_args_dict_to_cmd_list(cmd_args))
        for arg_name, value in cmd_args.items():
            assert getattr(args, arg_name) == value
        assert getattr(args, f"{prefix}url") == make_url(url)

    @pytest.mark.dependency(depends=["add_argument"])
    @pytest.mark.parametrize(
        "add_log_file, cmd_args, log_level",
        [
            param(False, {}, "WARNING", id="Default configuration"),
            param(False, {"log": "ERROR"}, "ERROR", id="ERROR log level"),
            param(False, {"debug": ""}, "DEBUG", id="DEBUG log level"),
            param(False, {"verbose": ""}, "INFO", id="INFO log level"),
            param(
                True,
                {"log_file": Path("app.log"), "log_file_level": "ERROR"},
                "WARNING",
                id="Add log file",
            ),
        ],
    )
    def test_add_log_arguments(self, add_log_file: bool, cmd_args: dict[str, Any], log_level: str) -> None:
        """Tests `ArgumentParser.add_log_arguments()` method.

        Args:
            add_log_file: Add arguments to allow storing messages into a file.
            cmd_args: Command line arguments to pass to the parser.
            log_level: Logging level expected.

        """
        # Add log arguments to the parser and its command line
        parser = ArgumentParser()
        parser.add_log_arguments(add_log_file=add_log_file)
        # Check that the argument is properly parsed
        args = parser.parse_args(_args_dict_to_cmd_list(cmd_args))
        assert args.log_level == log_level
        if add_log_file:
            for arg_name, value in cmd_args.items():
                assert getattr(args, arg_name) == value

    @pytest.mark.dependency(depends=["add_server_arguments"])
    def test_parse_args(self) -> None:
        """Tests `ArgumentParser.parse_args()` method.

        The only check left is when server arguments and a URL argument all with the same prefix
        are added to the parser.

        """
        # Add server arguments with prefix "src" and URL argument with the same prefix to the parser
        parser = ArgumentParser()
        parser.add_server_arguments("src_")
        parser.add_argument_url("--src_url")
        # Check that the error is handled properly
        with raises(SystemExit):
            parser.parse_args(["--src_host", "host", "--src_port", "42", "--src_user", "username"])



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.database.dbconnection` module."""

from contextlib import nullcontext as does_not_raise
import os
from typing import ContextManager

import pytest
from pytest import FixtureRequest, param, raises
from sqlalchemy import text
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.ext.automap import automap_base

from ensembl.utils.database import DBConnection, Query, UnitTestDB


@pytest.mark.parametrize("test_dbs", [[{"src": "mock_db"}]], indirect=True)
class TestDBConnection:
    """Tests `DBConnection` class.

    Attributes:
        dbc: Test database connection.
        server: Server URL where the test database is hosted.

    """

    dbc: DBConnection = None
    server: str = ""

    @pytest.fixture(scope="class", autouse=True)
    @classmethod
    def setup(cls, request: FixtureRequest, test_dbs: dict[str, UnitTestDB]) -> None:
        """Loads the required fixtures and values as class attributes.

        Args:
            request: Fixture that provides information of the requesting test function.
            test_dbs: Fixture that provides the unit test databases.

        """
        cls.dbc = test_dbs["mock_db"].dbc
        cls.server = request.config.getoption("server")

    @pytest.mark.dependency(name="test_init", scope="class")
    def test_init(self) -> None:
        """Tests that the object `DBConnection` is initialised correctly."""
        assert self.dbc, "DBConnection object should not be empty"

    @pytest.mark.dependency(name="test_dialect", depends=["test_init"], scope="class")
    def test_dialect(self) -> None:
        """Tests `DBConnection.dialect` property."""
        assert self.dbc.dialect == make_url(self.server).drivername

    @pytest.mark.dependency(name="test_db_name", depends=["test_init", "test_dialect"], scope="class")
    def test_db_name(self) -> None:
        """Tests `DBConnection.db_name` property."""
        expected_db_name = f"{os.environ['USER']}_mock_db"
        if self.dbc.dialect == "sqlite":
            expected_db_name += ".db"
        assert self.dbc.db_name == expected_db_name

    @pytest.mark.dependency(depends=["test_init", "test_dialect", "test_db_name"], scope="class")
    def test_url(self) -> None:
        """Tests `DBConnection.url` property."""
        expected_url = make_url(self.server).set(database=self.dbc.db_name)
        assert self.dbc.url == expected_url.render_as_string(hide_password=False)

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_host(self) -> None:
        """Tests `DBConnection.host` property."""
        assert self.dbc.host == make_url(self.server).host

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_port(self) -> None:
        """Tests `DBConnection.port` property."""
        assert self.dbc.port == make_url(self.server).port

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_tables(self) -> None:
        """Tests `DBConnection.tables` property."""
        assert set(self.dbc.tables.keys()) == {"gibberish"}

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_get_primary_key_columns(self) -> None:
        """Tests `DBConnection.get_primary_key_columns()` method."""
        table = "gibberish"
        assert set(self.dbc.get_primary_key_columns(table)) == {
            "id",
            "grp",
        }, f"Unexpected set of primary key columns found in table '{table}'"

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_get_columns(self) -> None:
        """Tests `DBConnection.get_columns()` method."""
        table = "gibberish"
        assert set(self.dbc.get_columns(table)) == {
            "id",
            "grp",
            "value",
        }, f"Unexpected set of columns found in table '{table}'"

    @pytest.mark.dependency(name="test_connect", depends=["test_init"], scope="class")
    def test_connect(self) -> None:
        """Tests `DBConnection.connect()` method."""
        connection = self.dbc.connect()
        assert connection, "Connection object should not be empty"
        result = connection.execute(text("SELECT * FROM gibberish"))
        assert len(result.fetchall()) == 6, "Unexpected number of rows found in 'gibberish' table"
        connection.close()

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_begin(self) -> None:
        """Tests `DBConnection.begin()` method."""
        with self.dbc.begin() as connection:
            assert connection, "Connection object should not be empty"
            result = connection.execute(text("SELECT * FROM gibberish"))
            assert len(result.fetchall()) == 6, "Unexpected number of rows found in 'gibberish' table"

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    def test_dispose(self) -> None:
        """Tests `DBConnection.dispose()` method."""
        self.dbc.dispose()
        # SQLAlchemy uses a "pool-less" connection system for SQLite
        if self.dbc.dialect != "sqlite":
            num_conn = self.dbc._engine.pool.checkedin()  # pylint: disable=protected-access
            assert num_conn == 0, "A new pool should have 0 checked-in connections"

    @pytest.mark.dependency(name="test_exec", depends=["test_init"], scope="class")
    @pytest.mark.parametrize(
        "query, nrows, expectation",
        [
            param("SELECT * FROM gibberish", 6, does_not_raise(), id="Valid string query"),
            param(text("SELECT * FROM gibberish"), 6, does_not_raise(), id="Valid text query"),
            param(
                "SELECT * FROM my_table",
                0,
                raises(SQLAlchemyError, match=r"(my_table.* doesn't exist|no such table: my_table)"),
                id="Querying an unexistent table",
            ),
        ],
    )
    def test_execute(self, query: Query, nrows: int, expectation: ContextManager) -> None:
        """Tests `DBConnection.execute()` method.

        Args:
            query: SQL query.
            nrows: Number of rows expected to be returned from the query.
            expectation: Context manager for the expected exception.

        """
        with expectation:
            result = self.dbc.execute(query)
            assert len(result.fetchall()) == nrows, "Unexpected number of rows returned"

    @pytest.mark.dependency(depends=["test_init", "test_connect", "test_exec"], scope="class")
    @pytest.mark.parametrize(
        "identifier, rows_to_add, before, after",
        [
            param(7, [{"grp": "grp4", "value": 1}, {"grp": "grp5", "value": 1}], 0, 2, id="Add new data"),
            param(
                7, [{"grp": "grp6", "value": 1}, {"grp": "grp6", "value": 2}], 2, 2, id="Add existing data"
            ),
        ],
    )
    def test_session_scope(
        self, identifier: int, rows_to_add: list[dict[str, str]], before: int, after: int
    ) -> None:
        """Tests `DBConnection.session_scope()` method.

        Bear in mind that the second parameterization of this test will fail if the dialect/table engine
        does not support rollback transactions.

        Args:
            identifier: ID of the rows to add.
            rows_to_add: Rows to add to the `gibberish` table.
            before: Number of rows in `gibberish` table for `id` before adding the rows.
            after: Number of rows in `gibberish` table for `id` after adding the rows.

        """
        query = f"SELECT * FROM gibberish WHERE id = {identifier}"
        results = self.dbc.execute(query)
        assert len(results.fetchall()) == before
        # Session requires mapped classes to interact with the database
        Base = automap_base()
        Base.prepare(autoload_with=self.dbc.connect())
        Gibberish = Base.classes.gibberish
        # Ignore IntegrityError raised when committing the new tags as some parametrizations will force it
        try:
            with self.dbc.session_scope() as session:
                rows = [Gibberish(id=identifier, **x) for x in rows_to_add]
                session.add_all(rows)
        except IntegrityError:
            pass
        results = self.dbc.execute(query)
        assert len(results.fetchall()) == after

    @pytest.mark.dependency(depends=["test_init", "test_connect", "test_exec"], scope="class")
    def test_test_session_scope(self) -> None:
        """Tests `DBConnection.test_session_scope()` method."""
        # Session requires mapped classes to interact with the database
        Base = automap_base()
        Base.prepare(autoload_with=self.dbc.connect())
        Gibberish = Base.classes.gibberish
        # Check that the tags added during the context manager are removed afterwards
        identifier = 8
        with self.dbc.test_session_scope() as session:
            results = session.query(Gibberish).filter_by(id=identifier)
            assert not results.all(), f"ID {identifier} should not have any entries"
            session.add(Gibberish(id=identifier, grp="grp7", value=15))
            session.add(Gibberish(id=identifier, grp="grp8", value=25))
            session.commit()
            results = session.query(Gibberish).filter_by(id=identifier)
            assert len(results.all()) == 2, f"ID {identifier} should have two rows"
        results = self.dbc.execute(f"SELECT * FROM gibberish WHERE id = {identifier}")
        if (
            self.dbc.dialect == "mysql"
            and self.dbc.tables["gibberish"].dialect_options["mysql"]["engine"] == "MyISAM"
        ):
            assert len(results.all()) == 2, f"SQLite/MyISAM: 2 rows permanently added to ID {identifier}"
        else:
            assert not results.fetchall(), f"No entries should have been permanently added to ID {identifier}"



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.logging` module."""

import argparse
import logging
from pathlib import Path

import pytest
from pytest import param

from ensembl.utils.logging import LogLevel, init_logging, init_logging_with_args


@pytest.mark.parametrize(
    "log_level, log_file",
    [
        param(logging.WARNING, "", id="No log file"),
        param(logging.ERROR, "app.log", id="Add log file"),
    ],
)
def test_init_logging(tmp_path: Path, log_level: LogLevel, log_file: str) -> None:
    """Tests `init_logging()` function.

    Args:
        tmp_path: Fixture that provides a temporary directory path unique to the test invocation.
        log_level: Minimum logging level.
        log_file: Logging file where to write logging messages besides the standard error.

    """
    if log_file:
        num_handlers = 2
        init_logging(log_level=log_level, log_file=tmp_path / log_file, log_file_level=log_level)
    else:
        num_handlers = 1
        init_logging(log_level=log_level)
    assert len(logging.root.handlers) == num_handlers
    for handler in logging.root.handlers:
        assert handler.level == log_level


def test_init_logging_with_args() -> None:
    """Tests `init_logging_with_args()` function."""
    parser = argparse.ArgumentParser()
    parser.add_argument("--log_level")
    args = parser.parse_args(["--log_level", "DEBUG"])
    init_logging_with_args(args)
    assert len(logging.root.handlers) == 1
    # NOTE: from 3.11 the log_level difference can be changed to: logging.getLevelNamesMapping().keys()
    assert logging.root.handlers[0].level == logging.DEBUG



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.rloader` module."""

import configparser
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

import pytest
from pytest import param, raises
import requests.exceptions

from ensembl.utils.rloader import RemoteFileLoader


@dataclass
class MockResponse:
    """Mocks `Response` object returned by `requests.get()` function."""

    text: str
    status_code: int = 200


def mock_requests_get(file_path: Path, *args, **kwargs) -> MockResponse:  # pylint: disable=unused-argument
    """Mocks `requests.get()` function, bypassing the required internet connection."""
    with file_path.open("r") as in_file:
        content = in_file.read()
    return MockResponse(text=content)


class TestRemoteFileLoader:
    """Tests `RemoteFileLoader` class."""

    @pytest.mark.parametrize(
        "file_format, output",
        [
            param("yaml", {"lang": "python", "os": ["linux", "windows"]}, id="YAML file"),
            param("json", {"lang": "python", "os": ["linux", "windows"]}, id="JSON file"),
            param("env", {"DEBUG": "True", "SECRET_KEY": "out_secret"}, id="ENV file"),
            param("txt", "lang: python\nsecret: yes", id="Unknown file format"),
        ],
    )
    @patch("requests.get")
    def test_r_open(self, mock_get: Mock, data_dir: Path, file_format: str, output: Any) -> None:
        """Tests `RemoteFileLoader.r_open()` method for different file formats.

        Args:
            mock_get: Fixture to mock the `requests.get()` function.
            data_dir: Fixture that provides the path to the test data folder matching the test's name.
            file_format: File format to select an ad-hoc parser.
            output: Expected returned output.

        """
        mock_get.side_effect = mock_requests_get
        loader = RemoteFileLoader(file_format)
        content = loader.r_open(data_dir / f"sample.{file_format}")
        assert content == output

    @patch("requests.get")
    def test_r_open_ini_format(self, mock_get: Mock, data_dir: Path) -> None:
        """Tests `RemoteFileLoader.r_open()` method for INI files.

        Args:
            mock_get: Fixture to mock the `requests.get()` function.
            data_dir: Fixture that provides the path to the test data folder matching the test's name.

        """
        mock_get.side_effect = mock_requests_get
        loader = RemoteFileLoader("ini")
        content = loader.r_open(data_dir / "sample.ini")
        parser = configparser.ConfigParser()
        parser["DEFAULT"] = {"DEBUG": "True", "SECRET_KEY": "out_secret"}
        assert content == parser

    @patch("requests.get")
    def test_r_open_status_code(self, mock_get: Mock) -> None:
        """Tests `RemoteFileLoader.r_open()` method when the status code returned is not 200.

        Args:
            mock_get: Fixture to mock the `requests.get()` function.

        """
        mock_get.return_value = MockResponse(text="", status_code=404)
        loader = RemoteFileLoader()
        with raises(requests.exceptions.HTTPError):
            loader.r_open("")

    @pytest.mark.parametrize(
        "exception",
        [
            param(requests.exceptions.HTTPError, id="HTTPError raised"),
            param(requests.exceptions.Timeout, id="Timeout raised"),
        ],
    )
    @patch("requests.get")
    def test_r_open_exceptions(self, mock_get: Mock, exception: Any) -> None:
        """Tests `RemoteFileLoader.r_open()` method when an exception is raised by the URL request.

        Args:
            mock_get: Fixture to mock the `requests.get()` function.
            exception: Expected exception raised.

        """
        mock_get.side_effect = exception
        loader = RemoteFileLoader()
        with raises(exception):
            loader.r_open("")



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing database handler.

This module provides the main class to create and drop testing databases, populating them from
preexisting dumps (if supplied).

Examples:

    >>> from ensembl.utils.database import UnitTestDB
    >>> test_db = UnitTestDB("mysql://user:passwd@mysql-server:4242/", "path/to/dumps", "my_db")
    >>> # You can access the database via test_db.dbc, for instance:
    >>> results = test_db.dbc.execute("SELECT * FROM my_table;")
    >>> # At the end do not forget to drop the database
    >>> test_db.drop()

"""

from __future__ import annotations

__all__ = [
    "UnitTestDB",
]

import os
from pathlib import Path
import subprocess
from typing import Optional

import sqlalchemy
from sqlalchemy import text
from sqlalchemy.engine import make_url
from sqlalchemy_utils.functions import create_database, database_exists, drop_database

from ensembl.utils import StrPath
from ensembl.utils.database import DBConnection, StrURL


class UnitTestDB:
    """Creates and connects to a new test database, applying the schema and importing the data.

    Args:
        server_url: URL of the server hosting the database.
        dump_dir: Directory path with the database schema in `table.sql` (mandatory) and one TSV data
            file (without headers) per table following the convention `<table_name>.txt` (optional).
        name: Name to give to the new database. If not provided, the last directory name of `dump_dir`
            will be used instead. In either case, the new database name will be prefixed by the username.

    Attributes:
        dbc: Database connection handler.

    Raises:
        FileNotFoundError: If `table.sql` is not found.

    """

    def __init__(self, server_url: StrURL, dump_dir: StrPath, name: Optional[str] = None) -> None:
        db_url = make_url(server_url)
        dump_dir_path = Path(dump_dir)
        db_name = os.environ["USER"] + "_" + (name if name else dump_dir_path.name)
        # Add the database name to the URL
        if db_url.get_dialect().name == "sqlite":
            db_url = db_url.set(database=f"{db_name}.db")
        else:
            db_url = db_url.set(database=db_name)
        # Enable "local_infile" variable for MySQL databases to allow importing data from files
        connect_args = {}
        if db_url.get_dialect().name == "mysql":
            connect_args["local_infile"] = 1
        # Create the database, dropping it beforehand if it already exists
        if database_exists(db_url):
            drop_database(db_url)
        create_database(db_url)
        # Establish the connection to the database, load the schema and import the data
        try:
            self.dbc = DBConnection(db_url, connect_args=connect_args)
            with self.dbc.begin() as conn:
                # Set InnoDB engine as default and disable foreign key checks for MySQL databases
                if self.dbc.dialect == "mysql":
                    conn.execute(text("SET default_storage_engine=InnoDB"))
                    conn.execute(text("SET FOREIGN_KEY_CHECKS=0"))
                # Load the schema
                with open(dump_dir_path / "table.sql", "r") as schema:
                    for query in "".join(schema.readlines()).split(";"):
                        if query.strip():
                            conn.execute(text(query))
                # And import any available data for each table
                for tsv_file in dump_dir_path.glob("*.txt"):
                    table = tsv_file.stem
                    self._load_data(conn, table, tsv_file)
                # Re-enable foreign key checks for MySQL databases
                if self.dbc.dialect == "mysql":
                    conn.execute(text("SET FOREIGN_KEY_CHECKS=1"))
        except:
            # Make sure the database is deleted before raising the exception
            drop_database(db_url)
            raise
        # Update the loaded metadata information of the database
        self.dbc.load_metadata()

    def __repr__(self) -> str:
        """Returns a string representation of this object."""
        return f"{self.__class__.__name__}({self.dbc.url!r})"

    def drop(self) -> None:
        """Drops the database."""
        drop_database(self.dbc.url)
        # Ensure the connection pool is properly closed and disposed
        self.dbc.dispose()

    def _load_data(self, conn: sqlalchemy.engine.Connection, table: str, src: StrPath) -> None:
        """Loads the table data from the given file.

        Args:
            conn: Open connection to the database.
            table: Table name to load the data to.
            src: File path with the data in TSV format (without headers).

        """
        if self.dbc.dialect == "sqlite":
            # SQLite does not have an equivalent to "LOAD DATA": use its ".import" command instead
            subprocess.run(["sqlite3", self.dbc.db_name, ".mode tabs", f".import {src} {table}"], check=True)
        elif self.dbc.dialect == "postgresql":
            conn.execute(text(f"COPY {table} FROM '{src}'"))
        elif self.dbc.dialect == "sqlserver":
            conn.execute(text(f"BULK INSERT {table} FROM '{src}'"))
        else:
            conn.execute(text(f"LOAD DATA LOCAL INFILE '{src}' INTO TABLE {table}"))



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Allow to seamlessly load / read the content of a remote file as if it was located locally."""

from __future__ import annotations

__all__ = ["RemoteFileLoader"]

import configparser
import json
import logging
from io import StringIO
from typing import Optional, Any

import dotenv
import requests
import requests.exceptions
import yaml


logger = logging.getLogger(__name__)


class RemoteFileLoader:
    """Loads remote files, allowing specific format parsing options.

    Args:
        parser: Parser to use for this object. Default: `None` (no format-specific parsing done).

    Attributes:
        available_formats: File formats with ad-hoc parsers available.
        parser: Parser selected for this object.

    """

    available_formats: set[str] = {"yaml", "ini", "env", "json"}
    parser: Optional[str] = None

    def __init__(self, parser: Optional[str] = None) -> None:
        if parser in self.available_formats:
            self.parser = parser

    def __parse(self, content: str) -> Any:
        if self.parser == "yaml":
            return yaml.load(content, yaml.SafeLoader)
        if self.parser == "ini":
            config = configparser.ConfigParser()
            try:
                config.read_string(content)
            except configparser.MissingSectionHeaderError:
                content = "[DEFAULT]\n" + content
                config.read_string(content)
            return config
        if self.parser == "env":
            return dotenv.dotenv_values(stream=StringIO(content))
        if self.parser == "json":
            return json.loads(content)
        # Only return content, no parsing
        return content

    def r_open(self, url: str) -> Any:
        """Returns the parsed remote file from the given URL.

        Args:
            url: URL of the remote file to fetch.

        Raises:
            requests.exception.HTTPError: If loading or requesting the given URL returned an error.
            requests.exception.Timeout: If a timeout was raised whilst requesting the given URL.

        """
        try:
            r = requests.get(url, timeout=120)
            if r.status_code == 200:
                return self.__parse(r.text)
            raise requests.exceptions.HTTPError(response=r)
        except requests.exceptions.HTTPError as ex:
            logger.exception(f"Error with request to {url}: {ex}")
            raise ex
        except requests.exceptions.Timeout as ex:
            logger.exception(f"Request timed out {url}: {ex}")
            raise ex



# See the NOTICE file distributed with this work for additional information
# regarding copyright ownership.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit testing of `ensembl.utils.database.unittestdb` module."""

from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import ContextManager, Optional

import pytest
from pytest import FixtureRequest, param, raises
from sqlalchemy_utils.functions import database_exists

from ensembl.utils.database import UnitTestDB


class TestUnitTestDB:
    """Tests `UnitTestDB` class.

    Attributes:
        dbs: Dictionary of `UnitTestDB` objects with the database name as key.

    """

    dbs: dict[str, UnitTestDB] = {}

    @pytest.mark.dependency(name="test_init", scope="class")
    @pytest.mark.parametrize(
        "src, name, expectation",
        [
            param(Path("mock_db"), None, does_not_raise(), id="Default test database creation"),
            param(Path("mock_db"), "renamed_db", does_not_raise(), id="Rename test database"),
            param(Path("mock_db"), None, does_not_raise(), id="Re-create mock db with absolute path"),
            param(Path("mock_dir"), None, raises(FileNotFoundError), id="Wrong dump folder"),
        ],
    )
    def test_init(
        self,
        request: FixtureRequest,
        data_dir: Path,
        src: Path,
        name: Optional[str],
        expectation: ContextManager,
    ) -> None:
        """Tests that the object `UnitTestDB` is initialised correctly.

        Args:
            request: Fixture that provides information of the requesting test function.
            data_dir: Fixture that provides the path to the test data folder matching the test's name.
            src: Directory path with the database schema and one TSV data file per table.
            name: Name to give to the new database.
            expectation: Context manager for the expected exception.

        """
        with expectation:
            server_url = request.config.getoption("server")
            src_path = src if src.is_absolute() else data_dir / src
            db_key = name if name else src.name
            self.dbs[db_key] = UnitTestDB(server_url, src_path, name)
            # Check that the database has been created correctly
            assert self.dbs[db_key], "UnitTestDB should not be empty"
            assert self.dbs[db_key].dbc, "UnitTestDB's database connection should not be empty"
            # Check that the database has been loaded correctly from the dump files
            result = self.dbs[db_key].dbc.execute("SELECT * FROM gibberish")
            assert len(result.fetchall()) == 6, "Unexpected number of rows found in 'gibberish' table"

    @pytest.mark.dependency(depends=["test_init"], scope="class")
    @pytest.mark.parametrize(
        "db_key",
        [
            param("mock_db"),
            param("renamed_db"),
        ],
    )
    def test_drop(self, db_key: str) -> None:
        """Tests the `UnitTestDB.drop()` method.

        Args:
            db_key: Key assigned to the UnitTestDB created in `TestUnitTestDB.test_init()`.

        """
        db_url = self.dbs[db_key].dbc.url
        assert database_exists(db_url)
        self.dbs[db_key].drop()
        assert not database_exists(db_url)
